#ifndef __KERNEL__
#  define __KERNEL__
#endif
#ifndef MODULE
#  define MODULE
#endif

#include <linux/module.h>

#include <linux/kernel.h>
#include <linux/malloc.h>
#include <linux/fs.h>
#include <linux/devfs_fs_kernel.h>
#include <linux/errno.h>
#include <linux/types.h>
#include <linux/stat.h>

#include <asm/uaccess.h>

#define DRUMS_MAJOR 60

static char *drums_strings[] = {"bam\n", "bum\n", "pam\n", "pum\n",
			       "tam\n", "tum\n"};
#define DRUMS_NR_DEV 6
#define DRUMS_NAME_LEN 3

ssize_t drums_read(struct file *filp , char *buf,
		 size_t count, loff_t *offp)
{
    struct inode *inode = filp->f_dentry->d_inode;
    int minor = MINOR(inode->i_rdev);
    char *txt;

    if (filp->private_data) {
	/* already assigned private_data: automatic numbers */
	txt = filp->private_data;
    } else {
	if (minor >= DRUMS_NR_DEV) return -ENODEV;
	txt = drums_strings[minor];
    }
    if (count > strlen(txt)) count = strlen(txt);
    copy_to_user(buf, txt, count);
    *offp += count;
    return count;
}


struct file_operations drums_fops = {
    read:    drums_read,
};

static devfs_handle_t drums_dir;
static devfs_handle_t drums_devs[DRUMS_NR_DEV];
static devfs_handle_t drums_tambourine, drums_timpani;

void cleanup_module(void)
{
    int i;
    for (i=0; i<DRUMS_NR_DEV; i++) {
	if (drums_devs[i])
	    devfs_unregister(drums_devs[i]);
    }
    if (drums_dir)
	devfs_unregister(drums_dir);
    if (drums_tambourine)
	devfs_unregister(drums_tambourine);
    if (drums_timpani)
	devfs_unregister(drums_timpani);
    devfs_unregister_chrdev(DRUMS_MAJOR, "drums");
}

int init_module(void)
{
    int i;
    /* old-style registration */
    i = devfs_register_chrdev(DRUMS_MAJOR, "drums", &drums_fops);
    if (i<0) {
	printk(KERN_WARNING "drums: can't register major %i\n", DRUMS_MAJOR);
	return i;
    }
    /* new stuff: create entry points in devfs */
    drums_dir = devfs_mk_dir(NULL, "drums", 0, NULL);
    if (!drums_dir) {
	printk(KERN_WARNING "drums: can't register /dev/drums\n");
	cleanup_module();
	return -EINVAL;
    }
    for (i=0; i<DRUMS_NR_DEV; i++) {
	drums_devs[i] = devfs_register(drums_dir,
				       drums_strings[i], DRUMS_NAME_LEN,
				       DEVFS_FL_NONE, DRUMS_MAJOR, i, 
				       S_IFCHR | S_IRUGO, 0, 0,
				      &drums_fops, NULL);
	if (!drums_devs[i]) {
	    printk(KERN_WARNING "drums: can't register /dev/drums/%s\n",
		   drums_strings[i]);
	    cleanup_module();
	    return -EINVAL;
	}
    }

    /* create the tambourine and timpani using dynamic numbers */
    drums_tambourine = devfs_register(NULL, "tambourine", 0,
				      DEVFS_FL_AUTO_DEVNUM, 0, 0,
				      S_IFCHR | S_IRUGO, 0, 0,
				      &drums_fops, (void *)"rattle\n");
    drums_timpani = devfs_register(NULL, "timpani", 0,
				   DEVFS_FL_AUTO_DEVNUM, 0, 0,
				   S_IFCHR | S_IRUGO, 0, 0,
				   &drums_fops, (void *)"boom\n");
    return 0;
}

